#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['primalscheme']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['biopython>=1,<2',
 'click==7.1.2',
 'parasail==1.2',
 'primer3-py>=0,<1',
 'progress==1.5',
 'reportlab>=3,<4']

extras_require = \
{'test': ['pytest >=2.7.3', 'pytest-cov']}

entry_points = \
{'console_scripts': ['primalscheme = primalscheme.cli:cli']}

setup(name='primalscheme',
      version='1.3.1',
      description='primalscheme is a tool for designing primer panels for multiplex PCR',
      author='Josh Quick',
      author_email='j.quick@bham.ac.uk',
      url='https://github.com/aresti/primalscheme',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6,<4',
     )
